// The MIT License
//
// Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Copyright (C) 2020 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import UIKit
import MaternityData

struct DataTableTitleAndValuesSet {
    let titleColumn: [DataTableCell]
    let valueColumns: [[DataTableCell]]
}

enum ObservationTableScanFetusMeasurementType: String, CaseIterable {
    case gestationAge
    case ulWeight
    case weightDeviation
    case fetusPresentation

    var title: String {
        switch self {
        case .gestationAge: return R.string.localizable.journal_section_scans_table_gestation_age()
        case .ulWeight: return R.string.localizable.journal_section_scans_table_ultrasound_weight()
        case .weightDeviation: return R.string.localizable.journal_section_scans_table_weight_deviation()
        case .fetusPresentation: return R.string.localizable.journal_section_scans_table_fetus_presentation()
        }
    }
}

enum ObservationTableScanMotherMeasurementType: String, CaseIterable {
    case gestationAge
    case amnioticFluid
    case cervix
    case dvp
    case conclusion

    var title: String {
        switch self {
        case .gestationAge: return R.string.localizable.journal_section_scans_table_gestation_age()
        case .amnioticFluid: return R.string.localizable.journal_section_scans_table_amniotic_fluid()
        case .cervix: return R.string.localizable.journal_section_scans_table_cervix()
        case .dvp: return R.string.localizable.journal_section_scans_table_dvp()
        case .conclusion: return R.string.localizable.journal_section_scans_table_conclusion()
        }
    }
}

extension ObservationTableScan {
    func toDataTableCells(dictionary: [String: String], titleWidth: CGFloat, valueWidth: CGFloat) -> DataTableTitleAndValuesSet {
        let titleColumn: [DataTableCell] = generateTitleColumn(dictionary: dictionary, width: titleWidth)
        let valueColumns: [[DataTableCell]] = generateValueColumns(width: valueWidth)
        return DataTableTitleAndValuesSet(
            titleColumn: titleColumn,
            valueColumns: valueColumns
        )
    }

    private func generateTitleColumn(dictionary: [String: String], width: CGFloat) -> [DataTableCell] {
        var titleColumn: [DataTableCell] = []
        for fetus in fetuses ?? [] {
            titleColumn.append(
                DataTableCell(
                    id: fetus.id ?? UUID().uuidString,
                    value: fetus.name ?? "-",
                    type: .boldText,
                    fixedWidth: width
                )
            )

            #warning("Dictionary implemenation is missing!")
            let fetusTitles = ObservationTableScanFetusMeasurementType.allCases.map { (measurementType) in
                DataTableCell(
                    id: measurementType.rawValue,
                    value: measurementType.title,
                    type: .text,
                    fixedWidth: width
                )
            }
            titleColumn.append(contentsOf: fetusTitles)
        }

        titleColumn.append(
            DataTableCell(
                id: "mother",
                value: R.string.localizable.appointment_details_measurement_mother(),
                type: .boldText,
                fixedWidth: width
            )
        )

        #warning("Dictionary implemenation is missing!")
        let motherTitles = ObservationTableScanMotherMeasurementType.allCases.map({ (measurementType) in
            DataTableCell(
                id: measurementType.rawValue,
                value: measurementType.title,
                type: .text,
                fixedWidth: width
            )
        })
        titleColumn.append(contentsOf: motherTitles)

        return titleColumn
    }

    private func generateValueColumns(width: CGFloat) -> [[DataTableCell]] {
        var valueColumns: [[DataTableCell]] = []

        let allFetusActivityIds: Set<String> = Set(
            [
                mother?.rows?.compactMap({ $0.activityId }) ?? [],
                fetuses?
                    .compactMap({ $0.rows?.compactMap({ $0.activityId }) })
                    .flatMap({ $0 })
                    .compactMap({ $0 }) ?? []
            ].flatMap({ $0 })
        )

        for activityId in allFetusActivityIds {
            var valueColumnForActivity: [DataTableCell] = []

            for fetus in fetuses ?? [] {
                let row = fetus.rows?.first(where: { $0.activityId == activityId })
                let activityDate = row?.date

                valueColumnForActivity.append(
                    DataTableCell(
                        id: "\(activityId)_date",
                        value: activityDate != nil ? Self.dateFormatter.string(from: activityDate!) : "-",
                        type: .boldText,
                        fixedWidth: width
                    )
                )

                let measurementTypes = ObservationTableScanFetusMeasurementType.allCases
                for measurementType in measurementTypes {
                    valueColumnForActivity.append(
                        dataTableFetusCell(for: measurementType, row: row, activityId: activityId, valueWidth: width)
                    )
                }
            }

            let row = mother?.rows?.first(where: { $0.activityId == activityId })
            let activityDate = row?.date
            valueColumnForActivity.append(
                DataTableCell(
                    id: "\(activityId)_date",
                    value: activityDate != nil ? Self.dateFormatter.string(from: activityDate!) : "-",
                    type: .boldText,
                    fixedWidth: width
                )
            )

            let measurementTypes = ObservationTableScanMotherMeasurementType.allCases
            for measurementType in measurementTypes {
                valueColumnForActivity.append(
                    dataTableMotherCell(for: measurementType, row: row, activityId: activityId, valueWidth: width)
                )
            }

            valueColumns.append(valueColumnForActivity)
        }
        return valueColumns
    }

    private func dataTableFetusCell(for measurementType: ObservationTableScanFetusMeasurementType, row: ObservationTableScanFetusRow?, activityId: String, valueWidth: CGFloat) -> DataTableCell {
        let measurementCellId = "\(activityId)_\(measurementType.rawValue)"
        let dataTableCell: DataTableCell
        switch measurementType {
        case .gestationAge:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.gestationAge)
        case .fetusPresentation:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.fetusPresentation)
        case .ulWeight:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.ulWeight?.description)
        case .weightDeviation:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.weightDeviation?.description)
        }
        return dataTableCell
    }

    private func dataTableMotherCell(for measurementType: ObservationTableScanMotherMeasurementType, row: ObservationTableScanMotherRow?, activityId: String, valueWidth: CGFloat) -> DataTableCell {
        let measurementCellId = "\(activityId)_\(measurementType.rawValue)"
        let dataTableCell: DataTableCell
        switch measurementType {
        case .gestationAge:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.gestationAge)
        case .amnioticFluid:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.amnioticFluid)
        case .cervix:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.cervix?.description)
        case .dvp:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.dvp?.description)
        case .conclusion:
            dataTableCell = dataTableValueCell(id: measurementCellId, valueWidth: valueWidth, value: row?.conclusion)
        }
        return dataTableCell
    }

    private func dataTableValueCell(id: String, valueWidth: CGFloat, value: String?) -> DataTableCell {
        DataTableCell(
            id: id,
            value: value ?? "-",
            type: .text,
            fixedWidth: valueWidth
        )
    }

    private static let dateFormatter: DateFormatter = {
        let df = DateFormatter()
        df.dateStyle = .short
        df.timeStyle = .none
        return df
    }()
}
